import json
import os
import re
import sys
import time
from datetime import datetime
from multiprocessing import Pool

import parse_citation
import psycopg2
import psycopg2.extras

QUERY_BATCH_SIZE = 200
PROCESS_COUNT = 12

def perform_insert_s2orc_one(input):
    cur = conn.cursor()
    cur.execute("""
        INSERT INTO s2orc_entry (id, arxivid, mag, acl, pubmed, pubmedcentral, dblp, doi, abstract, title) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
    """, input["entry_tuple"])

    psycopg2.extras.execute_batch(cur, "INSERT INTO s2orc_entry_author (paperid, author, count) VALUES (%s, %s, %s)", input["author_tuples"])
    psycopg2.extras.execute_batch(cur, "INSERT INTO s2orc_entry_bibauthor (paperid, bibauthor, count) VALUES (%s, %s, %s)", input["bibauthor_tuples"])
    psycopg2.extras.execute_batch(cur, "INSERT INTO s2orc_entry_bibentry (paperid, bibentry, count) VALUES (%s, %s, %s)", input["bibentry_tuples"])
    psycopg2.extras.execute_batch(cur, "INSERT INTO s2orc_entry_bibtitle (paperid, bibtitle, count) VALUES (%s, %s, %s)", input["bibtitle_tuples"])

    conn.commit()

def perform_insert_s2orc_bunch(inputs):
    cur = conn.cursor()
    psycopg2.extras.execute_batch(cur, "INSERT INTO s2orc_entry (id, arxivid, mag, acl, pubmed, pubmedcentral, dblp, doi, abstract, title) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)", [x["entry_tuple"] for x in inputs])

    psycopg2.extras.execute_batch(cur, "INSERT INTO s2orc_entry_author (paperid, author, count) VALUES (%s, %s, %s)", [x for inputv in inputs for x in inputv["author_tuples"]])
    psycopg2.extras.execute_batch(cur, "INSERT INTO s2orc_entry_bibauthor (paperid, bibauthor, count) VALUES (%s, %s, %s)", [x for inputv in inputs for x in inputv["bibauthor_tuples"]])
    psycopg2.extras.execute_batch(cur, "INSERT INTO s2orc_entry_bibentry (paperid, bibentry, count) VALUES (%s, %s, %s)", [x for inputv in inputs for x in inputv["bibentry_tuples"]])
    psycopg2.extras.execute_batch(cur, "INSERT INTO s2orc_entry_bibtitle (paperid, bibtitle, count) VALUES (%s, %s, %s)", [x for inputv in inputs for x in inputv["bibtitle_tuples"]])

    conn.commit()
    cur.close()

def perform_insert_author_bunch(inputs):
    cur = conn.cursor()
    psycopg2.extras.execute_batch(cur, "INSERT INTO author_entry (authorid, name, papercount, citationcount, hindex) VALUES (%s, %s, %s, %s, %s)", [x["author_tuple"] for x in inputs])
    conn.commit()
    cur.close()

def perform_insert_paper_bunch(inputs):
    cur = conn.cursor()

    psycopg2.extras.execute_batch(cur, "INSERT INTO paper_entry (corpusid, acl, dblp, arxiv, mag, pubmed, doi, pubmedcentral, url, title, year, referencecount, citationcount, isopenaccess) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)", [x["paper_tuple"] for x in inputs])

    psycopg2.extras.execute_batch(cur, "INSERT INTO paper_entry_paperauthor (corpusid, authorname, authorid) VALUES (%s, %s, %s)", [x for inputv in inputs for x in inputv["author_tuples"]])
    psycopg2.extras.execute_batch(cur, "INSERT INTO paper_entry_paperfield (corpusid, fieldofstudy, source) VALUES (%s, %s, %s)", [x for inputv in inputs for x in inputv["field_tuples"]])

    conn.commit()
    cur.close()

def perform_insert_citation_bunch(inputs):
    cur = conn.cursor()
    
    parse_citation.batch_insert_citation(cur, inputs)
    
    conn.commit()
    cur.close()

def perform_insert_bunch(inputs):
    if schema == "s2orc":
        perform_insert_s2orc_bunch(inputs)
    elif schema == "author":
        perform_insert_author_bunch(inputs)
    elif schema == "paper":
        perform_insert_paper_bunch(inputs)
    elif schema == "citation":
        perform_insert_citation_bunch(inputs)
        return
    else:
        print("This schema is not yet implemented")
        return

def process_input_s2orc(input):
    output = {}

    corpus_id = input["corpusid"]

    abstract = None
    if input["content"]["annotations"]["abstract"] is not None:
        abstract = [input["content"]["text"][int(x["start"]):int(x["end"])] for x in json.loads(input["content"]["annotations"]["abstract"])][0]

    title = None
    if input["content"]["annotations"]["title"] is not None:
        title = [input["content"]["text"][int(x["start"]):int(x["end"])] for x in json.loads(input["content"]["annotations"]["title"])][0]

    keys = ["arxiv", "mag", "acl", "pubmed", "pubmedcentral", "dblp", "doi"]

    if "externalids" not in input or input["externalids"] == None:
        input["externalids"] = {}

    for key in keys:
        if key not in input["externalids"]:
            input["externalids"][key] = None

    output["entry_tuple"] = (corpus_id, input["externalids"]["arxiv"], input["externalids"]["mag"], input["externalids"]["acl"], input["externalids"]["pubmed"], input["externalids"]["pubmedcentral"], input["externalids"]["dblp"], input["externalids"]["doi"], abstract, title)

    author_set = {}
    if input["content"]["annotations"]["author"] is not None:
        for author_idxs in json.loads(input["content"]["annotations"]["author"]):
            author_text = input["content"]["text"][int(author_idxs["start"]):int(author_idxs["end"])]
            re_match = re.match("([A-Za-z]+( [A-Za-z]+)+) ", author_text)
            if re_match is not None:
                author = str(re_match.group()).strip()
                if author not in author_set:
                    author_set[author] = 1
                else:
                    author_set[author] += 1
    output["author_tuples"] = [(corpus_id, author, author_set[author]) for author in author_set]

    bibauthor_set = {}
    if input["content"]["annotations"]["bibauthor"] is not None:
        for author_fragment in json.loads(input["content"]["annotations"]["bibauthor"]):
            author_text = input["content"]["text"][int(author_fragment["start"]):int(author_fragment["end"])]
            re_match = re.match("([a-zA-Z0-9 ]{2,})", author_text)
            if re_match is not None:
                bibauthor = str(re_match.group()).strip()
                if bibauthor not in bibauthor_set:
                    bibauthor_set[bibauthor] = 1
                else:
                    bibauthor_set[bibauthor] += 1
    output["bibauthor_tuples"] = [(corpus_id, bibauthor, bibauthor_set[bibauthor]) for bibauthor in bibauthor_set]

    bibentry_set = {}
    if input["content"]["annotations"]["bibentry"] is not None:
        for bibentry_idxs in json.loads(input["content"]["annotations"]["bibentry"]):
            bibentry = input["content"]["text"][int(bibentry_idxs["start"]):(int(bibentry_idxs["end"]))].strip()
            if bibentry not in bibentry_set:
                bibentry_set[bibentry] = 1
            else:
                bibentry_set[bibentry] += 1
    output["bibentry_tuples"] = [(corpus_id, bibentry, bibentry_set[bibentry]) for bibentry in bibentry_set]

    bibtitle_set = {}
    if input["content"]["annotations"]["bibtitle"] is not None:
        for bibtitle_idxs in json.loads(input["content"]["annotations"]["bibtitle"]):
            bibtitle = input["content"]["text"][int(bibtitle_idxs["start"]):(int(bibtitle_idxs["end"]))]
            if bibtitle not in bibtitle_set:
                bibtitle_set[bibtitle] = 1
            else:
                bibtitle_set[bibtitle] += 1
    output["bibtitle_tuples"] = [(corpus_id, bibtitle, bibtitle_set[bibtitle]) for bibtitle in bibtitle_set]

    return output

def process_input_author(input):
    return {
        "author_tuple": (input["authorid"], input["name"], input["papercount"], input["citationcount"], input["hindex"])
    }

def process_input_paper(input):
    if "externalids" not in input or input["externalids"] == None:
        input["externalids"] = {
            "ACL": None,
            "DBLP": None,
            "ArXiv": None,
            "MAG": None,
            "PubMed": None,
            "DOI": None,
            "PubMedCentral": None
        }
    
    corpus_id = input["corpusid"]
    
    paper_tuple = (
        corpus_id, 
        input["externalids"]["ACL"], 
        input["externalids"]["DBLP"], 
        input["externalids"]["ArXiv"], 
        input["externalids"]["MAG"], 
        input["externalids"]["PubMed"], 
        input["externalids"]["DOI"], 
        input["externalids"]["PubMedCentral"],
        input["url"],
        input["title"],
        int(input["year"]) if input["year"] != None else None,
        int(input["referencecount"]) if input["referencecount"] != None else None,
        int(input["citationcount"]) if input["citationcount"] != None else None,
        bool(input["isopenaccess"]) if input["isopenaccess"] != None else None
    )

    author_tuples = [(corpus_id, x["name"], x["authorId"]) for x in input["authors"]]
    
    if input["s2fieldsofstudy"] == None:
        input["s2fieldsofstudy"] = []

    field_tuples = [(corpus_id, x["category"], x["source"]) for x in input["s2fieldsofstudy"]]

    out = {
        "paper_tuple": paper_tuple,
        "author_tuples": author_tuples,
        "field_tuples": field_tuples
    }

    return out

def worker_init():
    global conn
    # conn = psycopg2.connect(<FILL THIS IN>)

def process_data_file_multi_threaded(file_name):
    entry_count = 0
    print("Processing file: {}".format(file_name))
    print("Multiprocessing mode: using {} processes.".format(PROCESS_COUNT))

    batched_inserts = []
    with Pool(PROCESS_COUNT, initializer=worker_init) as p:
        with open(file_name) as fin:
            for line in fin:
                entry_count += 1

                if max_entries != -1 and entry_count > max_entries:
                    break

                if len(batched_inserts) >= QUERY_BATCH_SIZE:
                    p.apply_async(perform_insert_bunch, args=([x for x in batched_inserts], ))
                    batched_inserts.clear()

                this_line = json.loads(line)
                if schema == "s2orc":
                    output = process_input_s2orc(this_line)
                    batched_inserts.append(output)
                elif schema == "author":
                    output = process_input_author(this_line)
                    batched_inserts.append(output)
                elif schema == "paper":
                    output = process_input_paper(this_line)
                    batched_inserts.append(output)
                elif schema == "citation":
                    base_tuple, context_tuples = parse_citation.process_line(this_line)
                    batched_inserts.append((base_tuple, context_tuples))
                else:
                    print("That schema is not yet implemented!")
                    break
                
        if len(batched_inserts) > 0:
            p.apply_async(perform_insert_bunch, args=([x for x in batched_inserts], ))
            batched_inserts.clear()
        p.close()
        p.join()
    return entry_count

def process_data_file_single_threaded(file_name):
    entry_count = 0
    print("Processing file: {}".format(file_name))
    print("Singleprocessing mode")

    worker_init()

    batched_inserts = []
    with open(file_name) as fin:
        for line in fin:
            entry_count += 1

            if max_entries != -1 and entry_count > max_entries:
                break

            if len(batched_inserts) >= QUERY_BATCH_SIZE:
                perform_insert_bunch(batched_inserts)
                batched_inserts.clear()

            this_line = json.loads(line)
            if schema == "s2orc":
                output = process_input_s2orc(this_line)
                batched_inserts.append(output)
            elif schema == "author":
                output = process_input_author(this_line)
                batched_inserts.append(output)
            elif schema == "paper":
                output = process_input_paper(this_line)
                batched_inserts.append(output)
            elif schema == "citation":
                base_tuple, context_tuples = parse_citation.process_line(this_line)
                batched_inserts.append((base_tuple, context_tuples))
            else:
                print("That schema is not yet implemented!")
                break
            
    if len(batched_inserts) > 0:
        perform_insert_bunch(batched_inserts)
        batched_inserts.clear()
    return entry_count

if __name__ == "__main__":

    if len(sys.argv) < 3:
        print("Usage: {} <file name> <file schema> [record limit]", os.argv[0])
    file_name = sys.argv[1]
    schema = sys.argv[2]

    if len(sys.argv) > 3:
        max_entries = int(sys.argv[3])
    else:
        max_entries = -1

    try:
        print("Starting processing of file", file_name)
        start_time = time.perf_counter_ns()
        entry_count = process_data_file_multi_threaded(file_name)
        end_time = time.perf_counter_ns()
        print("Time elapsed: {}".format(datetime.fromtimestamp((end_time - start_time) // 1000000000).strftime("%H:%M:%S")))
        print("Processed {} records, with an average speed of {} ms per record".format(entry_count, ((end_time - start_time) // 1000000) / entry_count))
    except Exception as e:
        print(e)
        raise e
