import json
import csv
import os
import random
import math
import os.path
from multiprocessing import Pool

import psycopg2

TOP_AUTHORS_DIR = "./top_authors/"
CACHE_DIR = "./cache/"
SUPERSTAR_COUNT = 3
SUPERSTAR_OFFSET = 5
PROCESS_COUNT = 8

# The sample count formula
# Chosen now as max(original_count, 2 + ceil(original_count * 0.025))
# This is arbitrarily chosen to show differences for more prolific authors in the graph without producing an unusable visualization
def sample_count(original_count):
    if original_count < 3:
        return original_count
    else:
        return 2 + math.ceil(0.025 * original_count)

def random_sample(list):
    return random.sample(list, sample_count(len(list)))

def load_top_authors(filename):
    out = []
    with open(TOP_AUTHORS_DIR + filename) as f:
        reader = csv.reader(f, delimiter=',')
        next(reader)
        for row in reader:
            if row[0] != "NULL" and row[1] != "NULL" and row[2] != "NULL":
                out.append({"authorid": int(row[0]), "authorname": row[1], "authorcitations": int(row[2])})
    return out

# Creates all edges from "red" to "green" nodes (red nodes are superstars, green nodes are associates)
def query_associated_authors(authorid):
    cache_path = CACHE_DIR + str(authorid) + "_associated_authors.json"
    out = []
    if os.path.isfile(cache_path):
        print("Using cached collaborators for {}".format(authorid))
        with open(cache_path) as f:
            out = json.load(f)
        return random_sample(out)
    print("Querying associated authors for {}".format(authorid))
    cur.execute("""WITH const (superstarid) AS (
            values (%s)
    )
    SELECT DISTINCT superstarpapers.title, superstarpapers.corpusid, associatedauthors.authorname, associatedauthors.authorid
        FROM paper_entry_paperauthor AS superstarpaperrefs
        JOIN paper_entry AS superstarpapers ON superstarpaperrefs.corpusid = superstarpapers.corpusid
        JOIN paper_entry_paperauthor AS associatedauthors ON superstarpapers.corpusid = associatedauthors.corpusid
        WHERE superstarpaperrefs.authorid = (SELECT superstarid FROM const)
                AND superstarpapers.year >= 2010 AND superstarpapers.year <= 2013
        AND associatedauthors.authorid IS NOT NULL
        AND associatedauthors.authorname IS NOT NULL
                AND associatedauthors.authorid != (SELECT superstarid FROM const);""",
    (authorid,))
    for row in cur.fetchall():
        out.append({"title": row[0], "corpusid": int(row[1]), "authorname": row[2], "authorid": int(row[3])})
    with open(cache_path, "w") as f:
        json.dump(out, f)
    return random_sample(out)

# Creates all edges from "green" to "grey" nodes (green nodes are associates, grey nodes are collaborators)
def query_collaborative_authors(authorid):
    cache_path = CACHE_DIR + str(authorid) + "_collaborative_authors.json"
    if os.path.isfile(cache_path):
        print("Using cached collaborators for {}".format(authorid))
        with open(cache_path) as f:
            out = json.load(f)
        return random_sample(out)
    print("Querying collaborative authors for {}".format(authorid))
    out = []
    cur.execute("""WITH const (associatedauthorid) AS (
            values (%s)
    )
    SELECT DISTINCT associatedauthorpapers.title, associatedauthorpapers.corpusid, collaborativeauthors.authorname, collaborativeauthors.authorid
        FROM paper_entry_paperauthor AS associatedauthorpaperrefs
        JOIN paper_entry AS associatedauthorpapers ON associatedauthorpaperrefs.corpusid = associatedauthorpapers.corpusid
        JOIN paper_entry_paperauthor AS collaborativeauthors ON associatedauthorpapers.corpusid = collaborativeauthors.corpusid
        WHERE associatedauthorpaperrefs.authorid = (SELECT associatedauthorid FROM const)
                AND associatedauthorpapers.year >= 2013 AND associatedauthorpapers.year <= 2023
        AND collaborativeauthors.authorid IS NOT NULL
        AND collaborativeauthors.authorname IS NOT NULL
                AND collaborativeauthors.authorid != (SELECT associatedauthorid FROM const);""",
    (authorid,))
    for row in cur.fetchall():
        out.append({"title": row[0], "corpusid": int(row[1]), "authorname": row[2], "authorid": int(row[3])})
    with open(cache_path, "w") as f:
        json.dump(out, f)
    return random_sample(out)

def populate_edge(point1, point2, data):
    if point1 < point2:
        if point1 not in edges:
            edges[point1] = {}
        if point2 not in edges[point1]:
            edges[point1][point2] = [data]
        else:
            edges[point1][point2].append(data)
    else:
        if point2 not in edges:
            edges[point2] = {}
        if point1 not in edges[point2]:
            edges[point2][point1] = [data]
        else:
            edges[point2][point1].append(data)

def worker_init():
    global conn, cur
    conn = psycopg2.connect(<FILL THIS IN>)
    cur = conn.cursor()

if __name__ == "__main__":
    conn = psycopg2.connect(<FILL THIS IN>)
    cur = conn.cursor()

    # Nodes store discipline and size
    nodes = {}
    # Edges are undirected and always indexed by the lower value node
    # Each dict (indexed by the lower value authorid) contains a dict of all of the higher value authorids it connects to, with values set to a list of tuples of paper ids and titles.
    # Example:
    """
    {
        1: {
            2: [(3, "The quick brown fox jumps over the lazy dog")]
        }
    }
    """
    edges = {}
    # Superstars by primary discipline
    superstars = {}

    with Pool(PROCESS_COUNT, initializer=worker_init) as p:
        files = [f for f in os.listdir(TOP_AUTHORS_DIR) if f.endswith(".csv")]

        for file in files:
            file_id = file.replace(".csv", "")
            superstars_tmp = load_top_authors(file)
            superstars[file_id] = [superstars_tmp[x] for x in range(0, SUPERSTAR_COUNT * SUPERSTAR_OFFSET - 1, SUPERSTAR_OFFSET)]

            for discipline in superstars:
                for superstar in superstars[discipline]:
                    author_id = superstar["authorid"]
                    print("Evaluating superstar {}".format(author_id))
                    nodes[author_id] = {
                        "discipline": discipline,
                        "citation_count": superstar["authorcitations"],
                        "author_name": superstar["authorname"],
                        "type": "superstar"
                    }
                    associated_authors = query_associated_authors(author_id)
                    print("There are {} authors associated with superstar {}".format(len(associated_authors), author_id))
                    collaborator_list = p.map(query_collaborative_authors, [x["authorid"] for x in associated_authors])
                    for (idx, associate) in enumerate(associated_authors):
                        associate_id = associate["authorid"]
                        # Populate the edge from superstar to associate (keeping our invariant: lesser ID first)
                        populate_edge(author_id, associate_id, (associate["corpusid"], associate["title"]))
                        # Populate the associate edge
                        if associate_id not in nodes:
                            nodes[associate_id] = {
                                "discipline": discipline,
                                "citation_count": -1,
                                "author_name": associate["authorname"],
                                "type": "associate"
                            }

                        for collaborator in collaborator_list[idx]:
                            collaborator_id = collaborator["authorid"]
                            populate_edge(associate_id, collaborator_id, (collaborator["corpusid"], collaborator["title"]))
                            if collaborator_id not in nodes:
                                nodes[collaborator_id] = {
                                    "discipline": discipline,
                                    "citation_count": -1,
                                    "author_name": collaborator["authorname"],
                                    "type": "collaborator"
                                }
    with open("superstars.json", "w") as f:
        json.dump(superstars, f)

    with open("graph.json", "w") as f:
        json.dump({"nodes": nodes, "edges": edges}, f)

    conn.commit()
    cur.close()
    conn.close()
