import json
import csv
import os
import random
import math
import os.path
from multiprocessing import Pool

import psycopg2

TOP_AUTHORS_DIR = "./top_authors/"
CACHE_DIR = "./cache/"
SUPERSTAR_COUNT = 3
SUPERSTAR_OFFSET = 5
PROCESS_COUNT = 8

def load_top_authors(filename):
    out = []
    with open(TOP_AUTHORS_DIR + filename) as f:
        reader = csv.reader(f, delimiter=',')
        next(reader)
        for row in reader:
            if row[0] != "NULL" and row[1] != "NULL" and row[2] != "NULL":
                out.append({"authorid": int(row[0]), "authorname": row[1], "authorcitations": int(row[2])})
    return out

# Creates all edges from "red" to "green" nodes (red nodes are superstars, green nodes are associates)
def query_associated_authors(authorid):
    cache_path = CACHE_DIR + str(authorid) + "_associated_authors.json"
    out = []
    if os.path.isfile(cache_path):
        print("Using cached collaborators for {}".format(authorid))
        with open(cache_path) as f:
            out = json.load(f)
        return out
    print("Querying associated authors for {}".format(authorid))
    cur.execute("""WITH const (superstarid) AS (
            values (%s)
    )
    SELECT DISTINCT superstarpapers.title, superstarpapers.corpusid, associatedauthors.authorname, associatedauthors.authorid
        FROM paper_entry_paperauthor AS superstarpaperrefs
        JOIN paper_entry AS superstarpapers ON superstarpaperrefs.corpusid = superstarpapers.corpusid
        JOIN paper_entry_paperauthor AS associatedauthors ON superstarpapers.corpusid = associatedauthors.corpusid
        WHERE superstarpaperrefs.authorid = (SELECT superstarid FROM const)
                AND superstarpapers.year >= 2010 AND superstarpapers.year <= 2013
        AND associatedauthors.authorid IS NOT NULL
        AND associatedauthors.authorname IS NOT NULL
                AND associatedauthors.authorid != (SELECT superstarid FROM const);""",
    (authorid,))
    for row in cur.fetchall():
        out.append({"title": row[0], "corpusid": int(row[1]), "authorname": row[2], "authorid": int(row[3])})
    with open(cache_path, "w") as f:
        json.dump(out, f)
    return out

def collaborator_count(authorid):
    cache_path = CACHE_DIR + str(authorid) + "_collaborative_authors.count"
    if os.path.isfile(cache_path):
        print("Using cached collaborators for {}".format(authorid))
        with open(cache_path) as f:
            out = int(f.read())
        return out
    print("Querying collaborative authors for {}".format(authorid))
    out = []
    cur.execute("""WITH const (associatedauthorid) AS (
            values (%s)
    )

    SELECT COUNT(DISTINCT collaborativeauthors.authorid)
        FROM paper_entry_paperauthor AS associatedauthorpaperrefs
        JOIN paper_entry AS associatedauthorpapers ON associatedauthorpaperrefs.corpusid = associatedauthorpapers.corpusid
        JOIN paper_entry_paperauthor AS collaborativeauthors ON associatedauthorpapers.corpusid = collaborativeauthors.corpusid
        WHERE associatedauthorpaperrefs.authorid = (SELECT associatedauthorid FROM const)
                AND associatedauthorpapers.year >= 2013 AND associatedauthorpapers.year <= 2023
        AND collaborativeauthors.authorid IS NOT NULL
        AND collaborativeauthors.authorname IS NOT NULL
                AND collaborativeauthors.authorid != (SELECT associatedauthorid FROM const);""",
    (authorid,))
    out = int(cur.fetchall()[0][0])
    with open(cache_path, "w") as f:
        f.write(str(out))
    return out

def worker_init():
    global conn, cur
    conn = psycopg2.connect(<FILL THIS IN>)
    cur = conn.cursor()

if __name__ == "__main__":
    conn = psycopg2.connect(<FILL THIS IN>)
    cur = conn.cursor()

    # Superstars by primary discipline
    superstars = {}

    total_associates = 0
    total_collaborators = 0

    with Pool(PROCESS_COUNT, initializer=worker_init) as p:
        files = [f for f in os.listdir(TOP_AUTHORS_DIR) if f.endswith(".csv")]

        for file in files:
            file_id = file.replace(".csv", "")
            superstars_tmp = load_top_authors(file)
            superstars[file_id] = [superstars_tmp[x] for x in range(0, SUPERSTAR_COUNT * SUPERSTAR_OFFSET - 1, SUPERSTAR_OFFSET)]

            for discipline in superstars:
                for superstar in superstars[discipline]:
                    author_id = superstar["authorid"]
                    print("Evaluating superstar {}".format(author_id))

                    associated_authors = query_associated_authors(author_id)
                    print("There are {} authors associated with superstar {}".format(len(associated_authors), author_id))
                    total_associates += len(associated_authors)
                    collaborator_counts = p.map(collaborator_count, [x["authorid"] for x in associated_authors])
                    total_collaborators += sum(collaborator_counts)

    print("There are {} associates and {} collaborators.".format(total_associates, total_collaborators))

    conn.commit()
    cur.close()
    conn.close()
